#   #!/usr/bin/env python
#   -*- coding: utf-8 -*-
#   ******************************************************************************
#     Copyright (c) 2024.
#     Developed by Yifei Lu
#     Last change on 3/26/24, 12:54 PM
#     Last change by yifei
#    *****************************************************************************

import numpy as np
import math

# Variables containing the common parameters in the GERG-2008 equations
RGERG = 8.314472
Rs = 8.31451
Rsr = Rs / RGERG
o13 = 1.0 / 3.0
NcGERG = 21
MaxFlds = 21
MaxMdl = 10
MaxTrmM = 12
MaxTrmP = 24
epsilon = 1e-15
coik = np.zeros((MaxFlds, MaxTrmP))
doik = np.zeros((MaxFlds, MaxTrmP))
dijk = np.zeros((MaxMdl, MaxTrmM))
Drold = 0
Trold = 0
Told = 0
Trold2 = 0
xold = np.zeros(MaxFlds)
mNumb = np.zeros((MaxFlds, MaxFlds))
kpol = np.zeros(MaxFlds)
kexp = np.zeros(MaxFlds)
kpolij = np.zeros(MaxMdl)
kexpij = np.zeros(MaxMdl)
Dc = np.zeros(MaxFlds)
Tc = np.zeros(MaxFlds)
MMiGERG = np.zeros(MaxFlds)
Vc3 = np.zeros(MaxFlds)
Tc2 = np.zeros(MaxFlds)
noik = np.zeros((MaxFlds, MaxTrmP))
toik = np.zeros((MaxFlds, MaxTrmP))
cijk = np.zeros((MaxMdl, MaxTrmM))
eijk = np.zeros((MaxMdl, MaxTrmM))
gijk = np.zeros((MaxMdl, MaxTrmM ))
nijk = np.zeros((MaxMdl, MaxTrmM ))
tijk = np.zeros((MaxMdl, MaxTrmM ))
btij = np.zeros((MaxFlds, MaxFlds))
bvij = np.zeros((MaxFlds, MaxFlds))
gtij = np.zeros((MaxFlds, MaxFlds))
gvij = np.zeros((MaxFlds, MaxFlds))
fij = np.zeros((MaxFlds, MaxFlds))
th0i = np.zeros((MaxFlds, 7))
n0i = np.zeros((MaxFlds, 7))
# taup = np.zeros((MaxFlds, MaxTrmP))
# taupijk = np.zeros((MaxFlds, MaxTrmM))
# Calculated in the PressureGERG subroutine, but not included as an argument since
# it is only used internally in the density algorithm.
# dPdDsave = 0

air_molar_mass = 28.97  # g/mol


# def SetupGERG():
#     """
#     Initialize all the constants and parameters in the GERG-2008 model.
#     Some values are modified for calculations that do not depend on T, D, and x in order to speed up the program.
#     :return:
#     """

gerg_gas_spices = np.array(['methane', 'nitrogen', 'carbon dioxide', 'ethane', 'propane', 'isobutane',
                            'butane', 'isopentane', 'pentane', 'hexane', 'heptane', 'octane', 'nonane',
                            'decane', 'hydrogen', 'oxygen', 'carbon monoxide', 'water', 'hydrogen sulfide',
                            'helium', 'argon'])

dict_components = {'methane': {'C': 1, 'H': 4},
                   'nitrogen': {'N': 2},
                   'carbon dioxide': {'C': 1, 'O': 2},
                   'ethane': {'C': 2, 'H': 6},
                   'propane': {'C': 3, 'H': 8},
                   'isobutane': {'C': 4, 'H': 10},
                   'n-butane': {'C': 4, 'H': 10},
                   'isopentane': {'C': 5, 'H': 12},
                   'n-pentane': {'C': 5, 'H': 12},
                   'n-hexane': {'C': 6, 'H': 14},
                   'n-heptane': {'C': 7, 'H': 16},
                   'n-octane': {'C': 8, 'H': 18},
                   'n-nonane': {'C': 9, 'H': 20},
                   'n-decane': {'C': 10, 'H': 22},
                   'hydrogen': {'H': 2},
                   'oxygen': {'O': 2},
                   'carbon monoxide': {'C': 1, 'O': 1},
                   'water': {'H': 2, 'O': 1},
                   'hydrogen sulfide': {'H': 2, 'S': 1},
                   'helium': {'He': 1},
                   'argon': {'Ar': 1}}

# Chemical elements ordered as C, H, N, O, S, He, Ar
chemical_elements = np.array(['C', 'H', 'N', 'O', 'S', 'He', 'Ar'])

# Initialize an empty 2D array
gerg_gas_chemical_composition = np.zeros((len(gerg_gas_spices), len(chemical_elements)), dtype=float)

# Fill in the array
for i, component in enumerate(gerg_gas_spices):
    for j, element in enumerate(chemical_elements):
        gerg_gas_chemical_composition[i, j] = dict_components.get(component, {}).get(element, 0)

gerg_gas_chemical_composition = np.ascontiguousarray(gerg_gas_chemical_composition.T)

bijk = np.zeros((MaxMdl, MaxTrmM))
T0 = 0
d0 = 0

for i in range(MaxFlds):
    xold[i] = 0

Told = 0

# Molar masses [g/mol]
MMiGERG[0] = 16.04246  # Methane
MMiGERG[1] = 28.0134  # Nitrogen
MMiGERG[2] = 44.0095  # Carbon dioxide
MMiGERG[3] = 30.06904  # Ethane
MMiGERG[4] = 44.09562  # Propane
MMiGERG[5] = 58.1222  # Isobutane
MMiGERG[6] = 58.1222  # n-Butane
MMiGERG[7] = 72.14878  # Isopentane
MMiGERG[8] = 72.14878  # n-Pentane
MMiGERG[9] = 86.17536  # Hexane
MMiGERG[10] = 100.20194  # Heptane
MMiGERG[11] = 114.22852  # Octane
MMiGERG[12] = 128.2551  # Nonane
MMiGERG[13] = 142.28168  # Decane
MMiGERG[14] = 2.01588  # Hydrogen
MMiGERG[15] = 31.9988  # Oxygen
MMiGERG[16] = 28.0101  # Carbon monoxide
MMiGERG[17] = 18.01528  # Water
MMiGERG[18] = 34.08088  # Hydrogen sulfide
MMiGERG[19] = 4.002602  # Helium
MMiGERG[20] = 39.948  # Argon

# Number of polynomial and exponential terms
for i in range(MaxFlds):
    kpol[i] = 6
    kexp[i] = 6

kexp[0] = 18
kexp[1] = 18
kexp[3] = 18
kpol[2] = 4
kexp[2] = 18
kpol[14] = 5
kexp[14] = 9
kpol[17] = 7
kexp[17] = 9
kpol[19] = 4
kexp[19] = 8
kpolij[0] = 2
kexpij[0] = 10
kpolij[1] = 5
kexpij[1] = 4
kpolij[2] = 2
kexpij[2] = 7
kpolij[3] = 3
kexpij[3] = 3
kpolij[4] = 2
kexpij[4] = 4
kpolij[5] = 3
kexpij[5] = 3
kpolij[6] = 4
kexpij[6] = 0
kpolij[9] = 10
kexpij[9] = 0

# Critical densities [mol/l]
Dc[0] = 10.139342719
Dc[1] = 11.1839
Dc[2] = 10.624978698
Dc[3] = 6.87085454
Dc[4] = 5.000043088
Dc[5] = 3.86014294
Dc[6] = 3.920016792
Dc[7] = 3.271
Dc[8] = 3.215577588
Dc[9] = 2.705877875
Dc[10] = 2.315324434
Dc[11] = 2.056404127
Dc[12] = 1.81
Dc[13] = 1.64
Dc[14] = 14.94
Dc[15] = 13.63
Dc[16] = 10.85
Dc[17] = 17.87371609
Dc[18] = 10.19
Dc[19] = 17.399
Dc[20] = 13.407429659

# Critical temperatures [K]
Tc[0] = 190.564
Tc[1] = 126.192
Tc[2] = 304.1282
Tc[3] = 305.322
Tc[4] = 369.825
Tc[5] = 407.817
Tc[6] = 425.125
Tc[7] = 460.35
Tc[8] = 469.7
Tc[9] = 507.82
Tc[10] = 540.13
Tc[11] = 569.32
Tc[12] = 594.55
Tc[13] = 617.7
Tc[14] = 33.19
Tc[15] = 154.595
Tc[16] = 132.86
Tc[17] = 647.096
Tc[18] = 373.1
Tc[19] = 5.1953
Tc[20] = 150.687

# Exponents in pure fluid equations
for i in range(MaxFlds):
    Vc3[i] = 1 / (Dc[i] ** o13) / 2
    Tc2[i] = math.sqrt(Tc[i])
    coik[i][0] = 0
    doik[i][0] = 1
    toik[i][0] = 0.25
    coik[i][1] = 0
    doik[i][1] = 1
    toik[i][1] = 1.125
    coik[i][2] = 0
    doik[i][2] = 1
    toik[i][2] = 1.5
    coik[i][3] = 0
    doik[i][3] = 2
    toik[i][3] = 1.375
    coik[i][4] = 0
    doik[i][4] = 3
    toik[i][4] = 0.25
    coik[i][5] = 0
    doik[i][5] = 7
    toik[i][5] = 0.875
    coik[i][6] = 1
    doik[i][6] = 2
    toik[i][6] = 0.625
    coik[i][7] = 1
    doik[i][7] = 5
    toik[i][7] = 1.75
    coik[i][8] = 2
    doik[i][8] = 1
    toik[i][8] = 3.625
    coik[i][9] = 2
    doik[i][9] = 4
    toik[i][9] = 3.625
    coik[i][10] = 3
    doik[i][10] = 3
    toik[i][10] = 14.5
    coik[i][11] = 3
    doik[i][11] = 4
    toik[i][11] = 12

for i in range(4):
    if i != 2:
        coik[i][0] = 0
        doik[i][0] = 1
        toik[i][0] = 0.125
        coik[i][1] = 0
        doik[i][1] = 1
        toik[i][1] = 1.125
        coik[i][2] = 0
        doik[i][2] = 2
        toik[i][2] = 0.375
        coik[i][3] = 0
        doik[i][3] = 2
        toik[i][3] = 1.125
        coik[i][4] = 0
        doik[i][4] = 4
        toik[i][4] = 0.625
        coik[i][5] = 0
        doik[i][5] = 4
        toik[i][5] = 1.5
        coik[i][6] = 1
        doik[i][6] = 1
        toik[i][6] = 0.625
        coik[i][7] = 1
        doik[i][7] = 1
        toik[i][7] = 2.625
        coik[i][8] = 1
        doik[i][8] = 1
        toik[i][8] = 2.75
        coik[i][9] = 1
        doik[i][9] = 2
        toik[i][9] = 2.125
        coik[i][10] = 1
        doik[i][10] = 3
        toik[i][10] = 2
        coik[i][11] = 1
        doik[i][11] = 6
        toik[i][11] = 1.75
        coik[i][12] = 2
        doik[i][12] = 2
        toik[i][12] = 4.5
        coik[i][13] = 2
        doik[i][13] = 3
        toik[i][13] = 4.75
        coik[i][14] = 2
        doik[i][14] = 3
        toik[i][14] = 5
        coik[i][15] = 2
        doik[i][15] = 4
        toik[i][15] = 4
        coik[i][16] = 2
        doik[i][16] = 4
        toik[i][16] = 4.5
        coik[i][17] = 3
        doik[i][17] = 2
        toik[i][17] = 7.5
        coik[i][18] = 3
        doik[i][18] = 3
        toik[i][18] = 14
        coik[i][19] = 3
        doik[i][19] = 4
        toik[i][19] = 11.5
        coik[i][20] = 6
        doik[i][20] = 5
        toik[i][20] = 26
        coik[i][21] = 6
        doik[i][21] = 6
        toik[i][21] = 28
        coik[i][22] = 6
        doik[i][22] = 6
        toik[i][22] = 30
        coik[i][23] = 6
        doik[i][23] = 7
        toik[i][23] = 16

# Coefficients of pure fluid equations
# Methane
noik[0][0] = 0.57335704239162
noik[0][1] = -1.676068752373
noik[0][2] = 0.23405291834916
noik[0][3] = -0.21947376343441
noik[0][4] = 0.016369201404128
noik[0][5] = 0.01500440638928
noik[0][6] = 0.098990489492918
noik[0][7] = 0.58382770929055
noik[0][8] = -0.7478686756039
noik[0][9] = 0.30033302857974
noik[0][10] = 0.20985543806568
noik[0][11] = -0.018590151133061
noik[0][12] = -0.15782558339049
noik[0][13] = 0.12716735220791
noik[0][14] = -0.032019743894346
noik[0][15] = -0.068049729364536
noik[0][16] = 0.024291412853736
noik[0][17] = 5.1440451639444E-03
noik[0][18] = -0.019084949733532
noik[0][19] = 5.5229677241291E-03
noik[0][20] = -4.4197392976085E-03
noik[0][21] = 0.040061416708429
noik[0][22] = -0.033752085907575
noik[0][23] = -2.5127658213357E-03
# Nitrogen
noik[1][0] = 0.59889711801201
noik[1][1] = -1.6941557480731
noik[1][2] = 0.24579736191718
noik[1][3] = -0.23722456755175
noik[1][4] = 0.017954918715141
noik[1][5] = 0.014592875720215
noik[1][6] = 0.10008065936206
noik[1][7] = 0.73157115385532
noik[1][8] = -0.88372272336366
noik[1][9] = 0.31887660246708
noik[1][10] = 0.20766491728799
noik[1][11] = -0.019379315454158
noik[1][12] = -0.16936641554983
noik[1][13] = 0.13546846041701
noik[1][14] = -0.033066712095307
noik[1][15] = -0.060690817018557
noik[1][16] = 0.012797548292871
noik[1][17] = 5.8743664107299E-03
noik[1][18] = -0.018451951971969
noik[1][19] = 4.7226622042472E-03
noik[1][20] = -5.2024079680599E-03
noik[1][21] = 0.043563505956635
noik[1][22] = -0.036251690750939
noik[1][23] = -2.8974026866543E-03
# Ethane
noik[3][0] = 0.63596780450714
noik[3][1] = -1.7377981785459
noik[3][2] = 0.28914060926272
noik[3][3] = -0.33714276845694
noik[3][4] = 0.022405964699561
noik[3][5] = 0.015715424886913
noik[3][6] = 0.11450634253745
noik[3][7] = 1.0612049379745
noik[3][8] = -1.2855224439423
noik[3][9] = 0.39414630777652
noik[3][10] = 0.31390924682041
noik[3][11] = -0.021592277117247
noik[3][12] = -0.21723666564905
noik[3][13] = -0.28999574439489
noik[3][14] = 0.42321173025732
noik[3][15] = 0.04643410025926
noik[3][16] = -0.13138398329741
noik[3][17] = 0.011492850364368
noik[3][18] = -0.033387688429909
noik[3][19] = 0.015183171583644
noik[3][20] = -4.7610805647657E-03
noik[3][21] = 0.046917166277885
noik[3][22] = -0.039401755804649
noik[3][23] = -3.2569956247611E-03
# Propane
noik[4][0] = 1.0403973107358
noik[4][1] = -2.8318404081403
noik[4][2] = 0.84393809606294
noik[4][3] = -0.076559591850023
noik[4][4] = 0.09469737305728
noik[4][5] = 2.4796475497006E-04
noik[4][6] = 0.2774376042287
noik[4][7] = -0.043846000648377
noik[4][8] = -0.2699106478435
noik[4][9] = -0.06931341308986
noik[4][10] = -0.029632145981653
noik[4][11] = 0.01404012675138
# Isobutane
noik[5][0] = 1.04293315891
noik[5][1] = -2.8184272548892
noik[5][2] = 0.8617623239785
noik[5][3] = -0.10613619452487
noik[5][4] = 0.098615749302134
noik[5][5] = 2.3948208682322E-04
noik[5][6] = 0.3033000485695
noik[5][7] = -0.041598156135099
noik[5][8] = -0.29991937470058
noik[5][9] = -0.080369342764109
noik[5][10] = -0.029761373251151
noik[5][11] = 0.01305963030314
# n-Butane
noik[6][0] = 1.0626277411455
noik[6][1] = -2.862095182835
noik[6][2] = 0.88738233403777
noik[6][3] = -0.12570581155345
noik[6][4] = 0.10286308708106
noik[6][5] = 2.5358040602654E-04
noik[6][6] = 0.32325200233982
noik[6][7] = -0.037950761057432
noik[6][8] = -0.32534802014452
noik[6][9] = -0.079050969051011
noik[6][10] = -0.020636720547775
noik[6][11] = 0.005705380933475
# Isopentane
noik[7][0] = 1.0963
noik[7][1] = -3.0402
noik[7][2] = 1.0317
noik[7][3] = -0.1541
noik[7][4] = 0.11535
noik[7][5] = 0.00029809
noik[7][6] = 0.39571
noik[7][7] = -0.045881
noik[7][8] = -0.35804
noik[7][9] = -0.10107
noik[7][10] = -0.035484
noik[7][11] = 0.018156
# n-Pentane
noik[8][0] = 1.0968643098001
noik[8][1] = -2.9988888298061
noik[8][2] = 0.99516886799212
noik[8][3] = -0.16170708558539
noik[8][4] = 0.11334460072775
noik[8][5] = 2.6760595150748E-04
noik[8][6] = 0.40979881986931
noik[8][7] = -0.040876423083075
noik[8][8] = -0.38169482469447
noik[8][9] = -0.10931956843993
noik[8][10] = -0.03207322332799
noik[8][11] = 0.016877016216975
# Hexane
noik[9][0] = 1.0553238013661
noik[9][1] = -2.6120615890629
noik[9][2] = 0.7661388296726
noik[9][3] = -0.29770320622459
noik[9][4] = 0.11879907733358
noik[9][5] = 2.7922861062617E-04
noik[9][6] = 0.46347589844105
noik[9][7] = 0.011433196980297
noik[9][8] = -0.48256968738131
noik[9][9] = -0.093750558924659
noik[9][10] = -6.7273247155994E-03
noik[9][11] = -5.1141583585428E-03
# Heptane
noik[10][0] = 1.0543747645262
noik[10][1] = -2.6500681506144
noik[10][2] = 0.81730047827543
noik[10][3] = -0.30451391253428
noik[10][4] = 0.122538687108
noik[10][5] = 2.7266472743928E-04
noik[10][6] = 0.4986582568167
noik[10][7] = -7.1432815084176E-04
noik[10][8] = -0.5423689552545
noik[10][9] = -0.13801821610756
noik[10][10] = -6.1595287380011E-03
noik[10][11] = 4.8602510393022E-04
# Octane
noik[11][0] = 1.0722544875633
noik[11][1] = -2.4632951172003
noik[11][2] = 0.65386674054928
noik[11][3] = -0.36324974085628
noik[11][4] = 0.12713269626764
noik[11][5] = 3.071357277793E-04
noik[11][6] = 0.5265685698754
noik[11][7] = 0.019362862857653
noik[11][8] = -0.58939426849155
noik[11][9] = -0.14069963991934
noik[11][10] = -7.8966330500036E-03
noik[11][11] = 3.3036597968109E-03
# Nonane
noik[12][0] = 1.1151
noik[12][1] = -2.702
noik[12][2] = 0.83416
noik[12][3] = -0.38828
noik[12][4] = 0.1376
noik[12][5] = 0.00028185
noik[12][6] = 0.62037
noik[12][7] = 0.015847
noik[12][8] = -0.61726
noik[12][9] = -0.15043
noik[12][10] = -0.012982
noik[12][11] = 0.0044325
# Decane
noik[13][0] = 1.0461
noik[13][1] = -2.4807
noik[13][2] = 0.74372
noik[13][3] = -0.52579
noik[13][4] = 0.15315
noik[13][5] = 0.00032865
noik[13][6] = 0.84178
noik[13][7] = 0.055424
noik[13][8] = -0.73555
noik[13][9] = -0.18507
noik[13][10] = -0.020775
noik[13][11] = 0.012335
# Oxygen
noik[15][0] = 0.88878286369701
noik[15][1] = -2.4879433312148
noik[15][2] = 0.59750190775886
noik[15][3] = 9.6501817061881E-03
noik[15][4] = 0.07197042871277
noik[15][5] = 2.2337443000195E-04
noik[15][6] = 0.18558686391474
noik[15][7] = -0.03812936803576
noik[15][8] = -0.15352245383006
noik[15][9] = -0.026726814910919
noik[15][10] = -0.025675298677127
noik[15][11] = 9.5714302123668E-03
# Carbon monoxide
noik[16][0] = 0.90554
noik[16][1] = -2.4515
noik[16][2] = 0.53149
noik[16][3] = 0.024173
noik[16][4] = 0.072156
noik[16][5] = 0.00018818
noik[16][6] = 0.19405
noik[16][7] = -0.043268
noik[16][8] = -0.12778
noik[16][9] = -0.027896
noik[16][10] = -0.034154
noik[16][11] = 0.016329
# Hydrogen sulfide
noik[18][0] = 0.87641
noik[18][1] = -2.0367
noik[18][2] = 0.21634
noik[18][3] = -0.050199
noik[18][4] = 0.066994
noik[18][5] = 0.00019076
noik[18][6] = 0.20227
noik[18][7] = -0.0045348
noik[18][8] = -0.2223
noik[18][9] = -0.034714
noik[18][10] = -0.014885
noik[18][11] = 0.0074154
# Argon
noik[20][0] = 0.85095714803969
noik[20][1] = -2.400322294348
noik[20][2] = 0.54127841476466
noik[20][3] = 0.016919770692538
noik[20][4] = 0.068825965019035
noik[20][5] = 2.1428032815338E-04
noik[20][6] = 0.17429895321992
noik[20][7] = -0.033654495604194
noik[20][8] = -0.13526799857691
noik[20][9] = -0.016387350791552
noik[20][10] = -0.024987666851475
noik[20][11] = 8.8769204815709E-03
# Carbon dioxide
coik[2][0] = 0
doik[2][0] = 1
toik[2][0] = 0
noik[2][0] = 0.52646564804653
coik[2][1] = 0
doik[2][1] = 1
toik[2][1] = 1.25
noik[2][1] = -1.4995725042592
coik[2][2] = 0
doik[2][2] = 2
toik[2][2] = 1.625
noik[2][2] = 0.27329786733782
coik[2][3] = 0
doik[2][3] = 3
toik[2][3] = 0.375
noik[2][3] = 0.12949500022786
coik[2][4] = 1
doik[2][4] = 3
toik[2][4] = 0.375
noik[2][4] = 0.15404088341841
coik[2][5] = 1
doik[2][5] = 3
toik[2][5] = 1.375
noik[2][5] = -0.58186950946814
coik[2][6] = 1
doik[2][6] = 4
toik[2][6] = 1.125
noik[2][6] = -0.18022494838296
coik[2][7] = 1
doik[2][7] = 5
toik[2][7] = 1.375
noik[2][7] = -0.095389904072812
coik[2][8] = 1
doik[2][8] = 6
toik[2][8] = 0.125
noik[2][8] = -8.0486819317679E-03
coik[2][9] = 1
doik[2][9] = 6
toik[2][9] = 1.625
noik[2][9] = -0.03554775127309
coik[2][10] = 2
doik[2][10] = 1
toik[2][10] = 3.75
noik[2][10] = -0.28079014882405
coik[2][11] = 2
doik[2][11] = 4
toik[2][11] = 3.5
noik[2][11] = -0.082435890081677
coik[2][12] = 3
doik[2][12] = 1
toik[2][12] = 7.5
noik[2][12] = 0.010832427979006
coik[2][13] = 3
doik[2][13] = 1
toik[2][13] = 8
noik[2][13] = -6.7073993161097E-03
coik[2][14] = 3
doik[2][14] = 3
toik[2][14] = 6
noik[2][14] = -4.6827907600524E-03
coik[2][15] = 3
doik[2][15] = 3
toik[2][15] = 16
noik[2][15] = -0.028359911832177
coik[2][16] = 3
doik[2][16] = 4
toik[2][16] = 11
noik[2][16] = 0.019500174744098
coik[2][17] = 5
doik[2][17] = 5
toik[2][17] = 24
noik[2][17] = -0.21609137507166
coik[2][18] = 5
doik[2][18] = 5
toik[2][18] = 26
noik[2][18] = 0.43772794926972
coik[2][19] = 5
doik[2][19] = 5
toik[2][19] = 28
noik[2][19] = -0.22130790113593
coik[2][20] = 6
doik[2][20] = 5
toik[2][20] = 24
noik[2][20] = 0.015190189957331
coik[2][21] = 6
doik[2][21] = 5
toik[2][21] = 26
noik[2][21] = -0.0153809489533
# Hydrogen
coik[14][0] = 0
doik[14][0] = 1
toik[14][0] = 0.5
noik[14][0] = 5.3579928451252
coik[14][1] = 0
doik[14][1] = 1
toik[14][1] = 0.625
noik[14][1] = -6.2050252530595
coik[14][2] = 0
doik[14][2] = 2
toik[14][2] = 0.375
noik[14][2] = 0.13830241327086
coik[14][3] = 0
doik[14][3] = 2
toik[14][3] = 0.625
noik[14][3] = -0.071397954896129
coik[14][4] = 0
doik[14][4] = 4
toik[14][4] = 1.125
noik[14][4] = 0.015474053959733
coik[14][5] = 1
doik[14][5] = 1
toik[14][5] = 2.625
noik[14][5] = -0.14976806405771
coik[14][6] = 1
doik[14][6] = 5
toik[14][6] = 0
noik[14][6] = -0.026368723988451
coik[14][7] = 1
doik[14][7] = 5
toik[14][7] = 0.25
noik[14][7] = 0.056681303156066
coik[14][8] = 1
doik[14][8] = 5
toik[14][8] = 1.375
noik[14][8] = -0.060063958030436
coik[14][9] = 2
doik[14][9] = 1
toik[14][9] = 4
noik[14][9] = -0.45043942027132
coik[14][10] = 2
doik[14][10] = 1
toik[14][10] = 4.25
noik[14][10] = 0.424788402445
coik[14][11] = 3
doik[14][11] = 2
toik[14][11] = 5
noik[14][11] = -0.021997640827139
coik[14][12] = 3
doik[14][12] = 5
toik[14][12] = 8
noik[14][12] = -0.01049952137453
coik[14][13] = 5
doik[14][13] = 1
toik[14][13] = 8
noik[14][13] = -2.8955902866816E-03
# Water
coik[17][0] = 0
doik[17][0] = 1
toik[17][0] = 0.5
noik[17][0] = 0.82728408749586
coik[17][1] = 0
doik[17][1] = 1
toik[17][1] = 1.25
noik[17][1] = -1.8602220416584
coik[17][2] = 0
doik[17][2] = 1
toik[17][2] = 1.875
noik[17][2] = -1.1199009613744
coik[17][3] = 0
doik[17][3] = 2
toik[17][3] = 0.125
noik[17][3] = 0.15635753976056
coik[17][4] = 0
doik[17][4] = 2
toik[17][4] = 1.5
noik[17][4] = 0.87375844859025
coik[17][5] = 0
doik[17][5] = 3
toik[17][5] = 1
noik[17][5] = -0.36674403715731
coik[17][6] = 0
doik[17][6] = 4
toik[17][6] = 0.75
noik[17][6] = 0.053987893432436
coik[17][7] = 1
doik[17][7] = 1
toik[17][7] = 1.5
noik[17][7] = 1.0957690214499
coik[17][8] = 1
doik[17][8] = 5
toik[17][8] = 0.625
noik[17][8] = 0.053213037828563
coik[17][9] = 1
doik[17][9] = 5
toik[17][9] = 2.625
noik[17][9] = 0.013050533930825
coik[17][10] = 2
doik[17][10] = 1
toik[17][10] = 5
noik[17][10] = -0.41079520434476
coik[17][11] = 2
doik[17][11] = 2
toik[17][11] = 4
noik[17][11] = 0.1463744334412
coik[17][12] = 2
doik[17][12] = 4
toik[17][12] = 4.5
noik[17][12] = -0.055726838623719
coik[17][13] = 3
doik[17][13] = 4
toik[17][13] = 3
noik[17][13] = -0.0112017741438
coik[17][14] = 5
doik[17][14] = 1
toik[17][14] = 4
noik[17][14] = -6.6062758068099E-03
coik[17][15] = 5
doik[17][15] = 1
toik[17][15] = 6
noik[17][15] = 4.6918522004538E-03
# Helium
coik[19][0] = 0
doik[19][0] = 1
toik[19][0] = 0
noik[19][0] = -0.45579024006737
coik[19][1] = 0
doik[19][1] = 1
toik[19][1] = 0.125
noik[19][1] = 1.2516390754925
coik[19][2] = 0
doik[19][2] = 1
toik[19][2] = 0.75
noik[19][2] = -1.5438231650621
coik[19][3] = 0
doik[19][3] = 4
toik[19][3] = 1
noik[19][3] = 0.020467489707221
coik[19][4] = 1
doik[19][4] = 1
toik[19][4] = 0.75
noik[19][4] = -0.34476212380781
coik[19][5] = 1
doik[19][5] = 3
toik[19][5] = 2.625
noik[19][5] = -0.020858459512787
coik[19][6] = 1
doik[19][6] = 5
toik[19][6] = 0.125
noik[19][6] = 0.016227414711778
coik[19][7] = 1
doik[19][7] = 5
toik[19][7] = 1.25
noik[19][7] = -0.057471818200892
coik[19][8] = 1
doik[19][8] = 5
toik[19][8] = 2
noik[19][8] = 0.019462416430715
coik[19][9] = 2
doik[19][9] = 2
toik[19][9] = 1
noik[19][9] = -0.03329568012302
coik[19][10] = 3
doik[19][10] = 1
toik[19][10] = 4.5
noik[19][10] = -0.010863577372367
coik[19][11] = 3
doik[19][11] = 2
toik[19][11] = 5
noik[19][11] = -0.022173365245954

# Exponents in mixture equations
# Methane-Nitrogen
dijk[2][0] = 1
tijk[2][0] = 0
cijk[2][0] = 0
eijk[2][0] = 0
bijk[2][0] = 0
gijk[2][0] = 0
nijk[2][0] = -9.8038985517335E-03
dijk[2][1] = 4
tijk[2][1] = 1.85
cijk[2][1] = 0
eijk[2][1] = 0
bijk[2][1] = 0
gijk[2][1] = 0
nijk[2][1] = 4.2487270143005E-04
dijk[2][2] = 1
tijk[2][2] = 7.85
cijk[2][2] = 1
eijk[2][2] = 0.5
bijk[2][2] = 1
gijk[2][2] = 0.5
nijk[2][2] = -0.034800214576142
dijk[2][3] = 2
tijk[2][3] = 5.4
cijk[2][3] = 1
eijk[2][3] = 0.5
bijk[2][3] = 1
gijk[2][3] = 0.5
nijk[2][3] = -0.13333813013896
dijk[2][4] = 2
tijk[2][4] = 0
cijk[2][4] = 0.25
eijk[2][4] = 0.5
bijk[2][4] = 2.5
gijk[2][4] = 0.5
nijk[2][4] = -0.011993694974627
dijk[2][5] = 2
tijk[2][5] = 0.75
cijk[2][5] = 0
eijk[2][5] = 0.5
bijk[2][5] = 3
gijk[2][5] = 0.5
nijk[2][5] = 0.069243379775168
dijk[2][6] = 2
tijk[2][6] = 2.8
cijk[2][6] = 0
eijk[2][6] = 0.5
bijk[2][6] = 3
gijk[2][6] = 0.5
nijk[2][6] = -0.31022508148249
dijk[2][7] = 2
tijk[2][7] = 4.45
cijk[2][7] = 0
eijk[2][7] = 0.5
bijk[2][7] = 3
gijk[2][7] = 0.5
nijk[2][7] = 0.24495491753226
dijk[2][8] = 3
tijk[2][8] = 4.25
cijk[2][8] = 0
eijk[2][8] = 0.5
bijk[2][8] = 3
gijk[2][8] = 0.5
nijk[2][8] = 0.22369816716981
# Methane-Carbon dioxide
dijk[3][0] = 1
tijk[3][0] = 2.6
cijk[3][0] = 0
eijk[3][0] = 0
bijk[3][0] = 0
gijk[3][0] = 0
nijk[3][0] = -0.10859387354942
dijk[3][1] = 2
tijk[3][1] = 1.95
cijk[3][1] = 0
eijk[3][1] = 0
bijk[3][1] = 0
gijk[3][1] = 0
nijk[3][1] = 0.080228576727389
dijk[3][2] = 3
tijk[3][2] = 0
cijk[3][2] = 0
eijk[3][2] = 0
bijk[3][2] = 0
gijk[3][2] = 0
nijk[3][2] = -9.3303985115717E-03
dijk[3][3] = 1
tijk[3][3] = 3.95
cijk[3][3] = 1
eijk[3][3] = 0.5
bijk[3][3] = 1
gijk[3][3] = 0.5
nijk[3][3] = 0.040989274005848
dijk[3][4] = 2
tijk[3][4] = 7.95
cijk[3][4] = 0.5
eijk[3][4] = 0.5
bijk[3][4] = 2
gijk[3][4] = 0.5
nijk[3][4] = -0.24338019772494
dijk[3][5] = 3
tijk[3][5] = 8
cijk[3][5] = 0
eijk[3][5] = 0.5
bijk[3][5] = 3
gijk[3][5] = 0.5
nijk[3][5] = 0.23855347281124
# Methane-Ethane
dijk[0][0] = 3
tijk[0][0] = 0.65
cijk[0][0] = 0
eijk[0][0] = 0
bijk[0][0] = 0
gijk[0][0] = 0
nijk[0][0] = -8.0926050298746E-04
dijk[0][1] = 4
tijk[0][1] = 1.55
cijk[0][1] = 0
eijk[0][1] = 0
bijk[0][1] = 0
gijk[0][1] = 0
nijk[0][1] = -7.5381925080059E-04
dijk[0][2] = 1
tijk[0][2] = 3.1
cijk[0][2] = 1
eijk[0][2] = 0.5
bijk[0][2] = 1
gijk[0][2] = 0.5
nijk[0][2] = -0.041618768891219
dijk[0][3] = 2
tijk[0][3] = 5.9
cijk[0][3] = 1
eijk[0][3] = 0.5
bijk[0][3] = 1
gijk[0][3] = 0.5
nijk[0][3] = -0.23452173681569
dijk[0][4] = 2
tijk[0][4] = 7.05
cijk[0][4] = 1
eijk[0][4] = 0.5
bijk[0][4] = 1
gijk[0][4] = 0.5
nijk[0][4] = 0.14003840584586
dijk[0][5] = 2
tijk[0][5] = 3.35
cijk[0][5] = 0.875
eijk[0][5] = 0.5
bijk[0][5] = 1.25
gijk[0][5] = 0.5
nijk[0][5] = 0.063281744807738
dijk[0][6] = 2
tijk[0][6] = 1.2
cijk[0][6] = 0.75
eijk[0][6] = 0.5
bijk[0][6] = 1.5
gijk[0][6] = 0.5
nijk[0][6] = -0.034660425848809
dijk[0][7] = 2
tijk[0][7] = 5.8
cijk[0][7] = 0.5
eijk[0][7] = 0.5
bijk[0][7] = 2
gijk[0][7] = 0.5
nijk[0][7] = -0.23918747334251
dijk[0][8] = 2
tijk[0][8] = 2.7
cijk[0][8] = 0
eijk[0][8] = 0.5
bijk[0][8] = 3
gijk[0][8] = 0.5
nijk[0][8] = 1.9855255066891E-03
dijk[0][9] = 3
tijk[0][9] = 0.45
cijk[0][9] = 0
eijk[0][9] = 0.5
bijk[0][9] = 3
gijk[0][9] = 0.5
nijk[0][9] = 6.1777746171555
dijk[0][10] = 3
tijk[0][10] = 0.55
cijk[0][10] = 0
eijk[0][10] = 0.5
bijk[0][10] = 3
gijk[0][10] = 0.5
nijk[0][10] = -6.9575358271105
dijk[0][11] = 3
tijk[0][11] = 1.95
cijk[0][11] = 0
eijk[0][11] = 0.5
bijk[0][11] = 3
gijk[0][11] = 0.5
nijk[0][11] = 1.0630185306388
# Methane-Propane
dijk[1][0] = 3
tijk[1][0] = 1.85
cijk[1][0] = 0
eijk[1][0] = 0
bijk[1][0] = 0
gijk[1][0] = 0
nijk[1][0] = 0.013746429958576
dijk[1][1] = 3
tijk[1][1] = 3.95
cijk[1][1] = 0
eijk[1][1] = 0
bijk[1][1] = 0
gijk[1][1] = 0
nijk[1][1] = -7.4425012129552E-03
dijk[1][2] = 4
tijk[1][2] = 0
cijk[1][2] = 0
eijk[1][2] = 0
bijk[1][2] = 0
gijk[1][2] = 0
nijk[1][2] = -4.5516600213685E-03
dijk[1][3] = 4
tijk[1][3] = 1.85
cijk[1][3] = 0
eijk[1][3] = 0
bijk[1][3] = 0
gijk[1][3] = 0
nijk[1][3] = -5.4546603350237E-03
dijk[1][4] = 4
tijk[1][4] = 3.85
cijk[1][4] = 0
eijk[1][4] = 0
bijk[1][4] = 0
gijk[1][4] = 0
nijk[1][4] = 2.3682016824471E-03
dijk[1][5] = 1
tijk[1][5] = 5.25
cijk[1][5] = 0.25
eijk[1][5] = 0.5
bijk[1][5] = 0.75
gijk[1][5] = 0.5
nijk[1][5] = 0.18007763721438
dijk[1][6] = 1
tijk[1][6] = 3.85
cijk[1][6] = 0.25
eijk[1][6] = 0.5
bijk[1][6] = 1
gijk[1][6] = 0.5
nijk[1][6] = -0.44773942932486
dijk[1][7] = 1
tijk[1][7] = 0.2
cijk[1][7] = 0
eijk[1][7] = 0.5
bijk[1][7] = 2
gijk[1][7] = 0.5
nijk[1][7] = 0.0193273748882
dijk[1][8] = 2
tijk[1][8] = 6.5
cijk[1][8] = 0
eijk[1][8] = 0.5
bijk[1][8] = 3
gijk[1][8] = 0.5
nijk[1][8] = -0.30632197804624
# Nitrogen-Carbon dioxide
dijk[4][0] = 2
tijk[4][0] = 1.85
cijk[4][0] = 0
eijk[4][0] = 0
bijk[4][0] = 0
gijk[4][0] = 0
nijk[4][0] = 0.28661625028399
dijk[4][1] = 3
tijk[4][1] = 1.4
cijk[4][1] = 0
eijk[4][1] = 0
bijk[4][1] = 0
gijk[4][1] = 0
nijk[4][1] = -0.10919833861247
dijk[4][2] = 1
tijk[4][2] = 3.2
cijk[4][2] = 0.25
eijk[4][2] = 0.5
bijk[4][2] = 0.75
gijk[4][2] = 0.5
nijk[4][2] = -1.137403208227
dijk[4][3] = 1
tijk[4][3] = 2.5
cijk[4][3] = 0.25
eijk[4][3] = 0.5
bijk[4][3] = 1
gijk[4][3] = 0.5
nijk[4][3] = 0.76580544237358
dijk[4][4] = 1
tijk[4][4] = 8
cijk[4][4] = 0
eijk[4][4] = 0.5
bijk[4][4] = 2
gijk[4][4] = 0.5
nijk[4][4] = 4.2638000926819E-03
dijk[4][5] = 2
tijk[4][5] = 3.75
cijk[4][5] = 0
eijk[4][5] = 0.5
bijk[4][5] = 3
gijk[4][5] = 0.5
nijk[4][5] = 0.17673538204534
# Nitrogen-Ethane
dijk[5][0] = 2
tijk[5][0] = 0
cijk[5][0] = 0
eijk[5][0] = 0
bijk[5][0] = 0
gijk[5][0] = 0
nijk[5][0] = -0.47376518126608
dijk[5][1] = 2
tijk[5][1] = 0.05
cijk[5][1] = 0
eijk[5][1] = 0
bijk[5][1] = 0
gijk[5][1] = 0
nijk[5][1] = 0.48961193461001
dijk[5][2] = 3
tijk[5][2] = 0
cijk[5][2] = 0
eijk[5][2] = 0
bijk[5][2] = 0
gijk[5][2] = 0
nijk[5][2] = -5.7011062090535E-03
dijk[5][3] = 1
tijk[5][3] = 3.65
cijk[5][3] = 1
eijk[5][3] = 0.5
bijk[5][3] = 1
gijk[5][3] = 0.5
nijk[5][3] = -0.1996682004132
dijk[5][4] = 2
tijk[5][4] = 4.9
cijk[5][4] = 1
eijk[5][4] = 0.5
bijk[5][4] = 1
gijk[5][4] = 0.5
nijk[5][4] = -0.69411103101723
dijk[5][5] = 2
tijk[5][5] = 4.45
cijk[5][5] = 0.875
eijk[5][5] = 0.5
bijk[5][5] = 1.25
gijk[5][5] = 0.5
nijk[5][5] = 0.69226192739021
# Methane-Hydrogen
dijk[6][0] = 1
tijk[6][0] = 2
cijk[6][0] = 0
eijk[6][0] = 0
bijk[6][0] = 0
gijk[6][0] = 0
nijk[6][0] = -0.25157134971934
dijk[6][1] = 3
tijk[6][1] = -1
cijk[6][1] = 0
eijk[6][1] = 0
bijk[6][1] = 0
gijk[6][1] = 0
nijk[6][1] = -6.2203841111983E-03
dijk[6][2] = 3
tijk[6][2] = 1.75
cijk[6][2] = 0
eijk[6][2] = 0
bijk[6][2] = 0
gijk[6][2] = 0
nijk[6][2] = 0.088850315184396

dijk[6][3] = 4
tijk[6][3] = 1.4
cijk[6][3] = 0
eijk[6][3] = 0
bijk[6][3] = 0
gijk[6][3] = 0
nijk[6][3] = -0.035592212573239
# Methane-n-Butane, Methane-Isobutane, Ethane-Propane, Ethane-n-Butane,
# Ethane-Isobutane, Propane-n-Butane, Propane-Isobutane, and n-Butane-Isobutane
dijk[9][0] = 1
tijk[9][0] = 1
cijk[9][0] = 0
eijk[9][0] = 0
bijk[9][0] = 0
gijk[9][0] = 0
nijk[9][0] = 2.5574776844118
dijk[9][1] = 1
tijk[9][1] = 1.55
cijk[9][1] = 0
eijk[9][1] = 0
bijk[9][1] = 0
gijk[9][1] = 0
nijk[9][1] = -7.9846357136353
dijk[9][2] = 1
tijk[9][2] = 1.7
cijk[9][2] = 0
eijk[9][2] = 0
bijk[9][2] = 0
gijk[9][2] = 0
nijk[9][2] = 4.7859131465806
dijk[9][3] = 2
tijk[9][3] = 0.25
cijk[9][3] = 0
eijk[9][3] = 0
bijk[9][3] = 0
gijk[9][3] = 0
nijk[9][3] = -0.73265392369587
dijk[9][4] = 2
tijk[9][4] = 1.35
cijk[9][4] = 0
eijk[9][4] = 0
bijk[9][4] = 0
gijk[9][4] = 0
nijk[9][4] = 1.3805471345312
dijk[9][5] = 3
tijk[9][5] = 0
cijk[9][5] = 0
eijk[9][5] = 0
bijk[9][5] = 0
gijk[9][5] = 0
nijk[9][5] = 0.28349603476365
dijk[9][6] = 3
tijk[9][6] = 1.25
cijk[9][6] = 0
eijk[9][6] = 0
bijk[9][6] = 0
gijk[9][6] = 0
nijk[9][6] = -0.49087385940425
dijk[9][7] = 4
tijk[9][7] = 0
cijk[9][7] = 0
eijk[9][7] = 0
bijk[9][7] = 0
gijk[9][7] = 0
nijk[9][7] = -0.10291888921447
dijk[9][8] = 4
tijk[9][8] = 0.7
cijk[9][8] = 0
eijk[9][8] = 0
bijk[9][8] = 0
gijk[9][8] = 0
nijk[9][8] = 0.11836314681968
dijk[9][9] = 4
tijk[9][9] = 5.4
cijk[9][9] = 0
eijk[9][9] = 0
bijk[9][9] = 0
gijk[9][9] = 0
nijk[9][9] = 5.5527385721943E-05

# Generalized parameters
fij[0][1] = 1  # Methane-Nitrogen
fij[0][2] = 1  # Methane-CO2
fij[0][3] = 1  # Methane-Ethane
fij[0][4] = 1  # Methane-Propane
fij[1][2] = 1  # Nitrogen-CO2
fij[1][3] = 1  # Nitrogen-Ethane
fij[0][14] = 1  # Methane-Hydrogen
fij[0][5] = 0.771035405688  # Methane-Isobutane
fij[0][6] = 1  # Methane-n-Butane
fij[3][4] = 0.13042476515  # Ethane-Propane
fij[3][5] = 0.260632376098  # Ethane-Isobutane
fij[3][6] = 0.281570073085  # Ethane-n-Butane
fij[4][5] = -0.0551609771024  # Propane-Isobutane
fij[4][6] = 0.0312572600489  # Propane-n-Butane
fij[5][6] = -0.0551240293009  # Isobutane-n-Butane

# Model numbers for binary mixtures with no excess functions (mn=-1)
for i in range(MaxFlds):
    mNumb[i][i] = -1
    for j in range(i + 1, MaxFlds):
        fij[j][i] = fij[i][j]
        mNumb[i][j] = -1
        mNumb[j][i] = -1

# Model numbers for excess functions, 10 is for generalized equation
mNumb[0][1] = 3
mNumb[0][2] = 4
mNumb[0][3] = 1
mNumb[0][4] = 2
mNumb[0][5] = 10
mNumb[0][6] = 10
mNumb[0][14] = 7
mNumb[1][2] = 5
mNumb[1][3] = 6
mNumb[3][4] = 10
mNumb[3][5] = 10
mNumb[3][6] = 10
mNumb[4][5] = 10
mNumb[4][6] = 10
mNumb[5][6] = 10

# Ideal gas parameters
n0i[0][2] = 4.00088
n0i[0][3] = 0.76315
n0i[0][4] = 0.0046
n0i[0][5] = 8.74432
n0i[0][6] = -4.46921
n0i[0][0] = 29.83843397
n0i[0][1] = -15999.69151
n0i[1][2] = 3.50031
n0i[1][3] = 0.13732
n0i[1][4] = -0.1466
n0i[1][5] = 0.90066
n0i[1][6] = 0
n0i[1][0] = 17.56770785
n0i[1][1] = -2801.729072
n0i[2][2] = 3.50002
n0i[2][3] = 2.04452
n0i[2][4] = -1.06044
n0i[2][5] = 2.03366
n0i[2][6] = 0.01393
n0i[2][0] = 20.65844696
n0i[2][1] = -4902.171516
n0i[3][2] = 4.00263
n0i[3][3] = 4.33939
n0i[3][4] = 1.23722
n0i[3][5] = 13.1974
n0i[3][6] = -6.01989
n0i[3][0] = 36.73005938
n0i[3][1] = -23639.65301
n0i[4][2] = 4.02939
n0i[4][3] = 6.60569
n0i[4][4] = 3.197
n0i[4][5] = 19.1921
n0i[4][6] = -8.37267
n0i[4][0] = 44.70909619
n0i[4][1] = -31236.63551
n0i[5][2] = 4.06714
n0i[5][3] = 8.97575
n0i[5][4] = 5.25156
n0i[5][5] = 25.1423
n0i[5][6] = 16.1388
n0i[5][0] = 34.30180349
n0i[5][1] = -38525.50276
n0i[6][2] = 4.33944
n0i[6][3] = 9.44893
n0i[6][4] = 6.89406
n0i[6][5] = 24.4618
n0i[6][6] = 14.7824
n0i[6][0] = 36.53237783
n0i[6][1] = -38957.80933
n0i[7][2] = 4
n0i[7][3] = 11.7618
n0i[7][4] = 20.1101
n0i[7][5] = 33.1688
n0i[7][6] = 0
n0i[7][0] = 43.17218626
n0i[7][1] = -51198.30946
n0i[8][2] = 4
n0i[8][3] = 8.95043
n0i[8][4] = 21.836
n0i[8][5] = 33.4032
n0i[8][6] = 0
n0i[8][0] = 42.67837089
n0i[8][1] = -45215.83
n0i[9][2] = 4
n0i[9][3] = 11.6977
n0i[9][4] = 26.8142
n0i[9][5] = 38.6164
n0i[9][6] = 0
n0i[9][0] = 46.99717188
n0i[9][1] = -52746.83318
n0i[10][2] = 4
n0i[10][3] = 13.7266
n0i[10][4] = 30.4707
n0i[10][5] = 43.5561
n0i[10][6] = 0
n0i[10][0] = 52.07631631
n0i[10][1] = -57104.81056
n0i[11][2] = 4
n0i[11][3] = 15.6865
n0i[11][4] = 33.8029
n0i[11][5] = 48.1731
n0i[11][6] = 0
n0i[11][0] = 57.25830934
n0i[11][1] = -60546.76385
n0i[12][2] = 4
n0i[12][3] = 18.0241
n0i[12][4] = 38.1235
n0i[12][5] = 53.3415
n0i[12][6] = 0
n0i[12][0] = 62.09646901
n0i[12][1] = -66600.12837
n0i[13][2] = 4
n0i[13][3] = 21.0069
n0i[13][4] = 43.4931
n0i[13][5] = 58.3657
n0i[13][6] = 0
n0i[13][0] = 65.93909154
n0i[13][1] = -74131.45483
n0i[14][2] = 2.47906
n0i[14][3] = 0.95806
n0i[14][4] = 0.45444
n0i[14][5] = 1.56039
n0i[14][6] = -1.3756
n0i[14][0] = 13.07520288
n0i[14][1] = -5836.943696
n0i[15][2] = 3.50146
n0i[15][3] = 1.07558
n0i[15][4] = 1.01334
n0i[15][5] = 0
n0i[15][6] = 0
n0i[15][0] = 16.8017173
n0i[15][1] = -2318.32269
n0i[16][2] = 3.50055
n0i[16][3] = 1.02865
n0i[16][4] = 0.00493
n0i[16][5] = 0
n0i[16][6] = 0
n0i[16][0] = 17.45786899
n0i[16][1] = -2635.244116
n0i[17][2] = 4.00392
n0i[17][3] = 0.01059
n0i[17][4] = 0.98763
n0i[17][5] = 3.06904
n0i[17][6] = 0
n0i[17][0] = 21.57882705
n0i[17][1] = -7766.733078
n0i[18][2] = 4
n0i[18][3] = 3.11942
n0i[18][4] = 1.00243
n0i[18][5] = 0
n0i[18][6] = 0
n0i[18][0] = 21.5830944
n0i[18][1] = -6069.035869
n0i[19][2] = 2.5
n0i[19][3] = 0
n0i[19][4] = 0
n0i[19][5] = 0
n0i[19][6] = 0
n0i[19][0] = 10.04639507
n0i[19][1] = -745.375
n0i[20][2] = 2.5
n0i[20][3] = 0
n0i[20][4] = 0
n0i[20][5] = 0
n0i[20][6] = 0
n0i[20][0] = 10.04639507
n0i[20][1] = -745.375
th0i[0][3] = 820.659
th0i[0][4] = 178.41
th0i[0][5] = 1062.82
th0i[0][6] = 1090.53
th0i[1][3] = 662.738
th0i[1][4] = 680.562
th0i[1][5] = 1740.06
th0i[1][6] = 0
th0i[2][3] = 919.306
th0i[2][4] = 865.07
th0i[2][5] = 483.553
th0i[2][6] = 341.109
th0i[3][3] = 559.314
th0i[3][4] = 223.284
th0i[3][5] = 1031.38
th0i[3][6] = 1071.29
th0i[4][3] = 479.856
th0i[4][4] = 200.893
th0i[4][5] = 955.312
th0i[4][6] = 1027.29
th0i[5][3] = 438.27
th0i[5][4] = 198.018
th0i[5][5] = 1905.02
th0i[5][6] = 893.765
th0i[6][3] = 468.27
th0i[6][4] = 183.636
th0i[6][5] = 1914.1
th0i[6][6] = 903.185
th0i[7][3] = 292.503
th0i[7][4] = 910.237
th0i[7][5] = 1919.37
th0i[7][6] = 0
th0i[8][3] = 178.67
th0i[8][4] = 840.538
th0i[8][5] = 1774.25
th0i[8][6] = 0
th0i[9][3] = 182.326
th0i[9][4] = 859.207
th0i[9][5] = 1826.59
th0i[9][6] = 0
th0i[10][3] = 169.789
th0i[10][4] = 836.195
th0i[10][5] = 1760.46
th0i[10][6] = 0
th0i[11][3] = 158.922
th0i[11][4] = 815.064
th0i[11][5] = 1693.07
th0i[11][6] = 0
th0i[12][3] = 156.854
th0i[12][4] = 814.882
th0i[12][5] = 1693.79
th0i[12][6] = 0
th0i[13][3] = 164.947
th0i[13][4] = 836.264
th0i[13][5] = 1750.24
th0i[13][6] = 0
th0i[14][3] = 228.734
th0i[14][4] = 326.843
th0i[14][5] = 1651.71
th0i[14][6] = 1671.69
th0i[15][3] = 2235.71
th0i[15][4] = 1116.69
th0i[15][5] = 0
th0i[15][6] = 0
th0i[16][3] = 1550.45
th0i[16][4] = 704.525
th0i[16][5] = 0
th0i[16][6] = 0
th0i[17][3] = 268.795
th0i[17][4] = 1141.41
th0i[17][5] = 2507.37
th0i[17][6] = 0
th0i[18][3] = 1833.63
th0i[18][4] = 847.181
th0i[18][5] = 0
th0i[18][6] = 0
th0i[19][3] = 0
th0i[19][4] = 0
th0i[19][5] = 0
th0i[19][6] = 0
th0i[20][3] = 0
th0i[20][4] = 0
th0i[20][5] = 0
th0i[20][6] = 0

# Mixture parameters for reducing variables
# CH4-N2
bvij[0][1] = 0.998721377
gvij[0][1] = 1.013950311
btij[0][1] = 0.99809883
gtij[0][1] = 0.979273013
# CH4-CO2
bvij[0][2] = 0.999518072
gvij[0][2] = 1.002806594
btij[0][2] = 1.02262449
gtij[0][2] = 0.975665369
# CH4-C2H6
bvij[0][3] = 0.997547866
gvij[0][3] = 1.006617867
btij[0][3] = 0.996336508
gtij[0][3] = 1.049707697
# CH4-C3H8
bvij[0][4] = 1.00482707
gvij[0][4] = 1.038470657
btij[0][4] = 0.989680305
gtij[0][4] = 1.098655531
# CH4-i-C4H10
bvij[0][5] = 1.011240388
gvij[0][5] = 1.054319053
btij[0][5] = 0.980315756
gtij[0][5] = 1.161117729
# CH4-C4H10
bvij[0][6] = 0.979105972
gvij[0][6] = 1.045375122
btij[0][6] = 0.99417491
gtij[0][6] = 1.171607691
# CH4-i-C5H12
bvij[0][7] = 1
gvij[0][7] = 1.343685343
btij[0][7] = 1
gtij[0][7] = 1.188899743
# CH4-C5H12
bvij[0][8] = 0.94833012
gvij[0][8] = 1.124508039
btij[0][8] = 0.992127525
gtij[0][8] = 1.249173968
# CH4-C6H14
bvij[0][9] = 0.958015294
gvij[0][9] = 1.052643846
btij[0][9] = 0.981844797
gtij[0][9] = 1.330570181
# CH4-C7H16
bvij[0][10] = 0.962050831
gvij[0][10] = 1.156655935
btij[0][10] = 0.977431529
gtij[0][10] = 1.379850328
# CH4-C8H18
bvij[0][11] = 0.994740603
gvij[0][11] = 1.116549372
btij[0][11] = 0.957473785
gtij[0][11] = 1.449245409
# CH4-C9H20
bvij[0][12] = 1.002852287
gvij[0][12] = 1.141895355
btij[0][12] = 0.947716769
gtij[0][12] = 1.528532478
# CH4-C10H22
bvij[0][13] = 1.033086292
gvij[0][13] = 1.146089637
btij[0][13] = 0.937777823
gtij[0][13] = 1.568231489
# CH4-H2
bvij[0][14] = 1
gvij[0][14] = 1.018702573
btij[0][14] = 1
gtij[0][14] = 1.352643115
# CH4-O2
bvij[0][15] = 1
gvij[0][15] = 1
btij[0][15] = 1
gtij[0][15] = 0.95
# CH4-CO
bvij[0][16] = 0.997340772
gvij[0][16] = 1.006102927
btij[0][16] = 0.987411732
gtij[0][16] = 0.987473033
# CH4-H2O
bvij[0][17] = 1.012783169
gvij[0][17] = 1.585018334
btij[0][17] = 1.063333913
gtij[0][17] = 0.775810513
# CH4-H2S
bvij[0][18] = 1.012599087
gvij[0][18] = 1.040161207
btij[0][18] = 1.011090031
gtij[0][18] = 0.961155729
# CH4-He
bvij[0][19] = 1
gvij[0][19] = 0.881405683
btij[0][19] = 1
gtij[0][19] = 3.159776855
# CH4-Ar
bvij[0][20] = 1.034630259
gvij[0][20] = 1.014678542
btij[0][20] = 0.990954281
gtij[0][20] = 0.989843388
# N2-CO2
bvij[1][2] = 0.977794634
gvij[1][2] = 1.047578256
btij[1][2] = 1.005894529
gtij[1][2] = 1.107654104
# N2-C2H6
bvij[1][3] = 0.978880168
gvij[1][3] = 1.042352891
btij[1][3] = 1.007671428
gtij[1][3] = 1.098650964
# N2-C3H8
bvij[1][4] = 0.974424681
gvij[1][4] = 1.081025408
btij[1][4] = 1.002677329
gtij[1][4] = 1.201264026
# N2-i-C4H10
bvij[1][5] = 0.98641583
gvij[1][5] = 1.100576129
btij[1][5] = 0.99286813
gtij[1][5] = 1.284462634
# N2-C4H10
bvij[1][6] = 0.99608261
gvij[1][6] = 1.146949309
btij[1][6] = 0.994515234
gtij[1][6] = 1.304886838
# N2-i-C5H12
bvij[1][7] = 1
gvij[1][7] = 1.154135439
btij[1][7] = 1
gtij[1][7] = 1.38177077
# N2-C5H12
bvij[1][8] = 1
gvij[1][8] = 1.078877166
btij[1][8] = 1
gtij[1][8] = 1.419029041
# N2-C6H14
bvij[1][9] = 1
gvij[1][9] = 1.195952177
btij[1][9] = 1
gtij[1][9] = 1.472607971
# N2-C7H16
bvij[1][10] = 1
gvij[1][10] = 1.40455409
btij[1][10] = 1
gtij[1][10] = 1.520975334
# N2-C8H18
bvij[1][11] = 1
gvij[1][11] = 1.186067025
btij[1][11] = 1
gtij[1][11] = 1.733280051
# N2-C9H20
bvij[1][12] = 1
gvij[1][12] = 1.100405929
btij[1][12] = 0.95637945
gtij[1][12] = 1.749119996
# N2-C10H22
bvij[1][13] = 1
gvij[1][13] = 1
btij[1][13] = 0.957934447
gtij[1][13] = 1.822157123
# N2-H2
bvij[1][14] = 0.972532065
gvij[1][14] = 0.970115357
btij[1][14] = 0.946134337
gtij[1][14] = 1.175696583
# N2-O2
bvij[1][15] = 0.99952177
gvij[1][15] = 0.997082328
btij[1][15] = 0.997190589
gtij[1][15] = 0.995157044
# N2-CO
bvij[1][16] = 1
gvij[1][16] = 1.008690943
btij[1][16] = 1
gtij[1][16] = 0.993425388

# N2-H2O
bvij[1][17] = 1
gvij[1][17] = 1.094749685
btij[1][17] = 1
gtij[1][17] = 0.968808467

# N2-H2S
bvij[1][18] = 0.910394249
gvij[1][18] = 1.256844157
btij[1][18] = 1.004692366
gtij[1][18] = 0.9601742

# N2-He
bvij[1][19] = 0.969501055
gvij[1][19] = 0.932629867
btij[1][19] = 0.692868765
gtij[1][19] = 1.47183158

# N2-Ar
bvij[1][20] = 1.004166412
gvij[1][20] = 1.002212182
btij[1][20] = 0.999069843
gtij[1][20] = 0.990034831

# CO2-C2H6
bvij[2][3] = 1.002525718
gvij[2][3] = 1.032876701
btij[2][3] = 1.013871147
gtij[2][3] = 0.90094953

# CO2-C3H8
bvij[2][4] = 0.996898004
gvij[2][4] = 1.047596298
btij[2][4] = 1.033620538
gtij[2][4] = 0.908772477

# CO2-i-C4H10
bvij[2][5] = 1.076551882
gvij[2][5] = 1.081909003
btij[2][5] = 1.023339824
gtij[2][5] = 0.929982936

# CO2-C4H10
bvij[2][6] = 1.174760923
gvij[2][6] = 1.222437324
btij[2][6] = 1.018171004
gtij[2][6] = 0.911498231

# CO2-i-C5H12
bvij[2][7] = 1.060793104
gvij[2][7] = 1.116793198
btij[2][7] = 1.019180957
gtij[2][7] = 0.961218039

# CO2-C5H12
bvij[2][8] = 1.024311498
gvij[2][8] = 1.068406078
btij[2][8] = 1.027000795
gtij[2][8] = 0.979217302

# CO2-C6H14
bvij[2][9] = 1
gvij[2][9] = 0.851343711
btij[2][9] = 1
gtij[2][9] = 1.038675574

# CO2-C7H16
bvij[2][10] = 1.205469976
gvij[2][10] = 1.164585914
btij[2][10] = 1.011806317
gtij[2][10] = 1.046169823

# CO2-C8H18
bvij[2][11] = 1.026169373
gvij[2][11] = 1.104043935
btij[2][11] = 1.02969078
gtij[2][11] = 1.074455386

# CO2-C9H20
bvij[2][12] = 1
gvij[2][12] = 0.973386152
btij[2][12] = 1.00768862
gtij[2][12] = 1.140671202

# CO2-C10H22
bvij[2][13] = 1.000151132
gvij[2][13] = 1.183394668
btij[2][13] = 1.02002879
gtij[2][13] = 1.145512213

# CO2-H2
bvij[2][14] = 0.904142159
gvij[2][14] = 1.15279255
btij[2][14] = 0.942320195
gtij[2][14] = 1.782924792

# CO2-O2
bvij[2][15] = 1
gvij[2][15] = 1
btij[2][15] = 1
gtij[2][15] = 1

# CO2-CO
bvij[2][16] = 1
gvij[2][16] = 1
btij[2][16] = 1
gtij[2][16] = 1

# CO2-H2O
bvij[2][17] = 0.949055959
gvij[2][17] = 1.542328793
btij[2][17] = 0.997372205
gtij[2][17] = 0.775453996

# CO2-H2S
bvij[2][18] = 0.906630564
gvij[2][18] = 1.024085837
btij[2][18] = 1.016034583
gtij[2][18] = 0.92601888

# CO2-He
bvij[2][19] = 0.846647561
gvij[2][19] = 0.864141549
btij[2][19] = 0.76837763
gtij[2][19] = 3.207456948

# CO2-Ar
bvij[2][20] = 1.008392428
gvij[2][20] = 1.029205465
btij[2][20] = 0.996512863
gtij[2][20] = 1.050971635

# C2H6-C3H8
bvij[3][4] = 0.997607277
gvij[3][4] = 1.00303472
btij[3][4] = 0.996199694
gtij[3][4] = 1.01473019

# C2H6-i-C4H10
bvij[3][5] = 1
gvij[3][5] = 1.006616886
btij[3][5] = 1
gtij[3][5] = 1.033283811

# C2H6-C4H10
bvij[3][6] = 0.999157205
gvij[3][6] = 1.006179146
btij[3][6] = 0.999130554
gtij[3][6] = 1.034832749

# C2H6-i-C5H12
bvij[3][7] = 1
gvij[3][7] = 1.045439935
btij[3][7] = 1
gtij[3][7] = 1.021150247

# C2H6-C5H12
bvij[3][8] = 0.993851009
gvij[3][8] = 1.026085655
btij[3][8] = 0.998688946
gtij[3][8] = 1.066665676

# C2H6-C6H14
bvij[3][9] = 1
gvij[3][9] = 1.169701102
btij[3][9] = 1
gtij[3][9] = 1.092177796

# C2H6-C7H16
bvij[3][10] = 1
gvij[3][10] = 1.057666085
btij[3][10] = 1
gtij[3][10] = 1.134532014

# C2H6-C8H18
bvij[3][11] = 1.007469726
gvij[3][11] = 1.071917985
btij[3][11] = 0.984068272
gtij[3][11] = 1.168636194

# C2H6-C9H20
bvij[3][12] = 1
gvij[3][12] = 1.14353473
btij[3][12] = 1
gtij[3][12] = 1.05603303

# C2H6-C10H22
bvij[3][13] = 0.995676258
gvij[3][13] = 1.098361281
btij[3][13] = 0.970918061
gtij[3][13] = 1.237191558

# C2H6-H2
bvij[3][14] = 0.925367171
gvij[3][14] = 1.10607204
btij[3][14] = 0.932969831
gtij[3][14] = 1.902008495

# C2H6-O2
bvij[3][15] = 1
gvij[3][15] = 1
btij[3][15] = 1
gtij[3][15] = 1

# C2H6-CO
bvij[3][16] = 1
gvij[3][16] = 1.201417898
btij[3][16] = 1
gtij[3][16] = 1.069224728

# C2H6-H2O
bvij[3][17] = 1
gvij[3][17] = 1
btij[3][17] = 1
gtij[3][17] = 1

# C2H6-H2S
bvij[3][18] = 1.010817909
gvij[3][18] = 1.030988277
btij[3][18] = 0.990197354
gtij[3][18] = 0.90273666

# C2H6-He
bvij[3][19] = 1
gvij[3][19] = 1
btij[3][19] = 1
gtij[3][19] = 1

# C2H6-Ar
bvij[3][20] = 1
gvij[3][20] = 1
btij[3][20] = 1
gtij[3][20] = 1

# C3H8-i-C4H10
bvij[4][5] = 0.999243146
gvij[4][5] = 1.001156119
btij[4][5] = 0.998012298
gtij[4][5] = 1.005250774

# C3H8-C4H10
bvij[4][6] = 0.999795868
gvij[4][6] = 1.003264179
btij[4][6] = 1.000310289
gtij[4][6] = 1.007392782

# C3H8-i-C5H12
bvij[4][7] = 1.040459289
gvij[4][7] = 0.999432118
btij[4][7] = 0.994364425
gtij[4][7] = 1.0032695

# C3H8-C5H12
bvij[4][8] = 1.044919431
gvij[4][8] = 1.019921513
btij[4][8] = 0.996484021
gtij[4][8] = 1.008344412

# C3H8-C6H14
bvij[4][9] = 1
gvij[4][9] = 1.057872566
btij[4][9] = 1
gtij[4][9] = 1.025657518

# C3H8-C7H16
bvij[4][10] = 1
gvij[4][10] = 1.079648053
btij[4][10] = 1
gtij[4][10] = 1.050044169

# C3H8-C8H18
bvij[4][11] = 1
gvij[4][11] = 1.102764612
btij[4][11] = 1
gtij[4][11] = 1.063694129

# C3H8-C9H20
bvij[4][12] = 1
gvij[4][12] = 1.199769134
btij[4][12] = 1
gtij[4][12] = 1.109973833

# C3H8-C10H22
bvij[4][13] = 0.984104227
gvij[4][13] = 1.053040574
btij[4][13] = 0.985331233
gtij[4][13] = 1.140905252

# C3H8-H2
bvij[4][14] = 1
gvij[4][14] = 1.07400611
btij[4][14] = 1
gtij[4][14] = 2.308215191

# C3H8-O2
bvij[4][15] = 1
gvij[4][15] = 1
btij[4][15] = 1
gtij[4][15] = 1

# C3H8-CO
bvij[4][16] = 1
gvij[4][16] = 1.108143673
btij[4][16] = 1
gtij[4][16] = 1.197564208

# C3H8-H2O
bvij[4][17] = 1
gvij[4][17] = 1.011759763
btij[4][17] = 1
gtij[4][17] = 0.600340961

# C3H8-H2S
bvij[4][18] = 0.936811219
gvij[4][18] = 1.010593999
btij[4][18] = 0.992573556
gtij[4][18] = 0.905829247

# C3H8-He
bvij[4][19] = 1
gvij[4][19] = 1
btij[4][19] = 1
gtij[4][19] = 1

# C3H8-Ar
bvij[4][20] = 1
gvij[4][20] = 1
btij[4][20] = 1
gtij[4][20] = 1

# The beta values for isobutane+butane are the reciprocal values of those in the GERG-2008 publication because the order was reversed in this work.
# C4H10 - i - C4H10
bvij[5][6] = 0.999120311
gvij[5][6] = 1.00041444
btij[5][6] = 0.999922459
gtij[5][6] = 1.001432824
# i-C4H10-i-C5H1
bvij[5][7] = 1
gvij[5][7] = 1.002284353
btij[5][7] = 1
gtij[5][7] = 1.001835788
# i-C4H10-C5H12
bvij[5][8] = 1
gvij[5][8] = 1.002779804
btij[5][8] = 1
gtij[5][8] = 1.002495889
# i - C4H10 - C6H14
bvij[5][9] = 1
gvij[5][9] = 1.010493989
btij[5][9] = 1
gtij[5][9] = 1.006018054
# i-C4H10-C7H16
bvij[5][10] = 1
gvij[5][10] = 1.021668316
btij[5][10] = 1
gtij[5][10] = 1.00988576
# i-C4H10-C8H18
bvij[5][11] = 1
gvij[5][11] = 1.032807063
btij[5][11] = 1
gtij[5][11] = 1.013945424
# i-C4H10-C9H20
bvij[5][12] = 1
gvij[5][12] = 1.047298475
btij[5][12] = 1
gtij[5][12] = 1.017817492
# i-C4H10-C10H22
bvij[5][13] = 1
gvij[5][13] = 1.060243344
btij[5][13] = 1
gtij[5][13] = 1.021624748
# i-C4H10-H2
bvij[5][14] = 1
gvij[5][14] = 1.147595688
btij[5][14] = 1
gtij[5][14] = 1.895305393
# i-C4H10-O2
bvij[5][15] = 1
gvij[5][15] = 1
btij[5][15] = 1
gtij[5][15] = 1
# i-C4H10-CO
bvij[5][16] = 1
gvij[5][16] = 1.087272232
btij[5][16] = 1
gtij[5][16] = 1.161390082
# i-C4H10-H2O
bvij[5][17] = 1
gvij[5][17] = 1
btij[5][17] = 1
gtij[5][17] = 1
# i-C4H10-H2S
bvij[5][18] = 1.012994431
gvij[5][18] = 0.988591117
btij[5][18] = 0.974550548
gtij[5][18] = 0.937130844
# i-C4H10-He
bvij[5][19] = 1
gvij[5][19] = 1
btij[5][19] = 1
gtij[5][19] = 1
# i-C4H10-Ar
bvij[5][20] = 1
gvij[5][20] = 1
btij[5][20] = 1
gtij[5][20] = 1
# C4H10-i-C5H12
bvij[6][7] = 1
gvij[6][7] = 1.002728434
btij[6][7] = 1
gtij[6][7] = 1.000792201
# C4H10-C5H12
bvij[6][8] = 1
gvij[6][8] = 1.01815965
btij[6][8] = 1
gtij[6][8] = 1.00214364
# C4H10-C6H14
bvij[6][9] = 1
gvij[6][9] = 1.034995284
btij[6][9] = 1
gtij[6][9] = 1.00915706
# C4H10-C7H16
bvij[6][10] = 1
gvij[6][10] = 1.019174227
btij[6][10] = 1
gtij[6][10] = 1.021283378
# C4H10-C8H18
bvij[6][11] = 1
gvij[6][11] = 1.046905515
btij[6][11] = 1
gtij[6][11] = 1.033180106
# C4H10-C9H20
bvij[6][12] = 1
gvij[6][12] = 1.049219137
btij[6][12] = 1
gtij[6][12] = 1.014096448
# C4H10-C10H22
bvij[6][13] = 0.976951968
gvij[6][13] = 1.027845529
btij[6][13] = 0.993688386
gtij[6][13] = 1.076466918
# C4H10-H2
bvij[6][14] = 1
gvij[6][14] = 1.232939523
btij[6][14] = 1
gtij[6][14] = 2.509259945
# C4H10-O2
bvij[6][15] = 1
gvij[6][15] = 1
btij[6][15] = 1
gtij[6][15] = 1
# C4H10-CO
bvij[6][16] = 1
gvij[6][16] = 1.084740904
btij[6][16] = 1
gtij[6][16] = 1.173916162
# C4H10-H2O
bvij[6][17] = 1
gvij[6][17] = 1.223638763
btij[6][17] = 1
gtij[6][17] = 0.615512682
# C4H10-H2S
bvij[6][18] = 0.908113163
gvij[6][18] = 1.033366041
btij[6][18] = 0.985962886
gtij[6][18] = 0.926156602
# C4H10-He
bvij[6][19] = 1
gvij[6][19] = 1
btij[6][19] = 1
gtij[6][19] = 1
# C4H10-Ar
bvij[6][20] = 1
gvij[6][20] = 1.214638734
btij[6][20] = 1
gtij[6][20] = 1.245039498
# C5H12-i-C5H12
bvij[7][8] = 1
gvij[7][8] = 1.000024335
btij[7][8] = 1
gtij[7][8] = 1.000050537
# i-C5H12-C6H14
bvij[7][9] = 1
gvij[7][9] = 1.002995876
btij[7][9] = 1
gtij[7][9] = 1.001204174
# i-C5H12-C7H16
bvij[7][10] = 1
gvij[7][10] = 1.009928206
btij[7][10] = 1
gtij[7][10] = 1.003194615
# i-C5H12-C8H18
bvij[7][11] = 1
gvij[7][11] = 1.017880545
btij[7][11] = 1
gtij[7][11] = 1.00564748
# i-C5H12-C9H20
bvij[7][12] = 1
gvij[7][12] = 1.028994325
btij[7][12] = 1
gtij[7][12] = 1.008191499
# i-C5H12-C10H22
bvij[7][13] = 1
gvij[7][13] = 1.039372957
btij[7][13] = 1
gtij[7][13] = 1.010825138
# i-C5H12-H2
bvij[7][14] = 1
gvij[7][14] = 1.184340443
btij[7][14] = 1
gtij[7][14] = 1.996386669
# i-C5H12-O2
bvij[7][15] = 1
gvij[7][15] = 1
btij[7][15] = 1
gtij[7][15] = 1
# i-C5H12-CO
bvij[7][16] = 1
gvij[7][16] = 1.116694577
btij[7][16] = 1
gtij[7][16] = 1.199326059
# i-C5H12-H2O
bvij[7][17] = 1
gvij[7][17] = 1
btij[7][17] = 1
gtij[7][17] = 1
# i-C5H12-H2S
bvij[7][18] = 1
gvij[7][18] = 0.835763343
btij[7][18] = 1
gtij[7][18] = 0.982651529
# i-C5H12-He
bvij[7][19] = 1
gvij[7][19] = 1
btij[7][19] = 1
gtij[7][19] = 1
# i-C5H12-Ar
bvij[7][20] = 1
gvij[7][20] = 1
btij[7][20] = 1
gtij[7][20] = 1
# C5H12-C6H14
bvij[8][9] = 1
gvij[8][9] = 1.002480637
btij[8][9] = 1
gtij[8][9] = 1.000761237
# C5H12-C7H16
bvij[8][10] = 1
gvij[8][10] = 1.008972412
btij[8][10] = 1
gtij[8][10] = 1.002441051
# C5H12-C8H18
bvij[8][11] = 1
gvij[8][11] = 1.069223964
btij[8][11] = 1
gtij[8][11] = 1.016422347
# C5H12-C9H20
bvij[8][12] = 1
gvij[8][12] = 1.034910633
btij[8][12] = 1
gtij[8][12] = 1.103421755
# C5H12-C10H22
bvij[8][13] = 1
gvij[8][13] = 1.016370338
btij[8][13] = 1
gtij[8][13] = 1.049035838
# C5H12-H2
bvij[8][14] = 1
gvij[8][14] = 1.188334783
btij[8][14] = 1
gtij[8][14] = 2.013859174
# C5H12-O2
bvij[8][15] = 1
gvij[8][15] = 1
btij[8][15] = 1
gtij[8][15] = 1
# C5H12-CO
bvij[8][16] = 1
gvij[8][16] = 1.119954454
btij[8][16] = 1
gtij[8][16] = 1.206043295

# C5H12-H2O
bvij[8][17] = 1
gvij[8][17] = 0.95667731
btij[8][17] = 1
gtij[8][17] = 0.447666011

# C5H12-H2S
bvij[8][18] = 0.984613203
gvij[8][18] = 1.076539234
btij[8][18] = 0.962006651
gtij[8][18] = 0.959065662

# C5H12-He
bvij[8][19] = 1
gvij[8][19] = 1
btij[8][19] = 1
gtij[8][19] = 1

# C5H12-Ar
bvij[8][20] = 1
gvij[8][20] = 1
btij[8][20] = 1
gtij[8][20] = 1

# C6H14-C7H16
bvij[9][10] = 1
gvij[9][10] = 1.001508227
btij[9][10] = 1
gtij[9][10] = 0.999762786

# C6H14-C8H18
bvij[9][11] = 1
gvij[9][11] = 1.006268954
btij[9][11] = 1
gtij[9][11] = 1.001633952

# C6H14-C9H20
bvij[9][12] = 1
gvij[9][12] = 1.02076168
btij[9][12] = 1
gtij[9][12] = 1.055369591

# C6H14-C10H22
bvij[9][13] = 1.001516371
gvij[9][13] = 1.013511439
btij[9][13] = 0.99764101
gtij[9][13] = 1.028939539

# C6H14-H2
bvij[9][14] = 1
gvij[9][14] = 1.243461678
btij[9][14] = 1
gtij[9][14] = 3.021197546

# C6H14-O2
bvij[9][15] = 1
gvij[9][15] = 1
btij[9][15] = 1
gtij[9][15] = 1

# C6H14-CO
bvij[9][16] = 1
gvij[9][16] = 1.155145836
btij[9][16] = 1
gtij[9][16] = 1.233272781

# C6H14-H2O
bvij[9][17] = 1
gvij[9][17] = 1.170217596
btij[9][17] = 1
gtij[9][17] = 0.569681333

# C6H14-H2S
bvij[9][18] = 0.754473958
gvij[9][18] = 1.339283552
btij[9][18] = 0.985891113
gtij[9][18] = 0.956075596

# C6H14-He
bvij[9][19] = 1
gvij[9][19] = 1
btij[9][19] = 1
gtij[9][19] = 1

# C6H14-Ar
bvij[9][20] = 1
gvij[9][20] = 1
btij[9][20] = 1
gtij[9][20] = 1

# C7H16-C8H18
bvij[10][11] = 1
gvij[10][11] = 1.006767176
btij[10][11] = 1
gtij[10][11] = 0.998793111

# C7H16-C9H20
bvij[10][12] = 1
gvij[10][12] = 1.001370076
btij[10][12] = 1
gtij[10][12] = 1.001150096

# C7H16-C10H22
bvij[10][13] = 1
gvij[10][13] = 1.002972346
btij[10][13] = 1
gtij[10][13] = 1.002229938

# C7H16-H2
bvij[10][14] = 1
gvij[10][14] = 1.159131722
btij[10][14] = 1
gtij[10][14] = 3.169143057

# C7H16-O2
bvij[10][15] = 1
gvij[10][15] = 1
btij[10][15] = 1
gtij[10][15] = 1

# C7H16-CO
bvij[10][16] = 1
gvij[10][16] = 1.190354273
btij[10][16] = 1
gtij[10][16] = 1.256123503

# C7H16-H2O
bvij[10][17] = 1
gvij[10][17] = 1
btij[10][17] = 1
gtij[10][17] = 1

# C7H16-H2S
bvij[10][18] = 0.828967164
gvij[10][18] = 1.087956749
btij[10][18] = 0.988937417
gtij[10][18] = 1.013453092

# C7H16-He
bvij[10][19] = 1
gvij[10][19] = 1
btij[10][19] = 1
gtij[10][19] = 1

# C7H16-Ar
bvij[10][20] = 1
gvij[10][20] = 1
btij[10][20] = 1
gtij[10][20] = 1

# C8H18-C9H20
bvij[11][12] = 1
gvij[11][12] = 1.001357085
btij[11][12] = 1
gtij[11][12] = 1.000235044

# C8H18-C10H22
bvij[11][13] = 1
gvij[11][13] = 1.002553544
btij[11][13] = 1
gtij[11][13] = 1.007186267

# C8H18-H2
bvij[11][14] = 1
gvij[11][14] = 1.305249405
btij[11][14] = 1
gtij[11][14] = 2.191555216

# C8H18-O2
bvij[11][15] = 1
gvij[11][15] = 1
btij[11][15] = 1
gtij[11][15] = 1

# C8H18-CO
bvij[11][16] = 1
gvij[11][16] = 1.219206702
btij[11][16] = 1
gtij[11][16] = 1.276565536

# C8H18-H2O
bvij[11][17] = 1
gvij[11][17] = 0.599484191
btij[11][17] = 1
gtij[11][17] = 0.662072469

# C8H18-H2S
bvij[11][18] = 1
gvij[11][18] = 1
btij[11][18] = 1
gtij[11][18] = 1

# C8H18-He
bvij[11][19] = 1
gvij[11][19] = 1
btij[11][19] = 1
gtij[11][19] = 1

# C8H18-Ar
bvij[11][20] = 1
gvij[11][20] = 1
btij[11][20] = 1
gtij[11][20] = 1

# C9H20-C10H22
bvij[12][13] = 1
gvij[12][13] = 1.00081052
btij[12][13] = 1
gtij[12][13] = 1.000182392

# C9H20-H2
bvij[12][14] = 1
gvij[12][14] = 1.342647661
btij[12][14] = 1
gtij[12][14] = 2.23435404

# C9H20-O2
bvij[12][15] = 1
gvij[12][15] = 1
btij[12][15] = 1
gtij[12][15] = 1

# C9H20-CO
bvij[12][16] = 1
gvij[12][16] = 1.252151449
btij[12][16] = 1
gtij[12][16] = 1.294070556

# C9H20-H2O
bvij[12][17] = 1
gvij[12][17] = 1
btij[12][17] = 1
gtij[12][17] = 1

# C9H20-H2S
bvij[12][18] = 1
gvij[12][18] = 1.082905109
btij[12][18] = 1
gtij[12][18] = 1.086557826

# C9H20-He
bvij[12][19] = 1
gvij[12][19] = 1
btij[12][19] = 1
gtij[12][19] = 1

# C9H20-Ar
bvij[12][20] = 1
gvij[12][20] = 1
btij[12][20] = 1
gtij[12][20] = 1

# C10H22-H2
bvij[13][14] = 1.695358382
gvij[13][14] = 1.120233729
btij[13][14] = 1.064818089
gtij[13][14] = 3.786003724

# C10H22-O2
bvij[13][15] = 1
gvij[13][15] = 1
btij[13][15] = 1
gtij[13][15] = 1

# C10H22-CO
bvij[13][16] = 1
gvij[13][16] = 0.87018496
btij[13][16] = 1.049594632
gtij[13][16] = 1.803567587

# C10H22-H2O
bvij[13][17] = 1
gvij[13][17] = 0.551405318
btij[13][17] = 0.897162268
gtij[13][17] = 0.740416402

# C10H22-H2S
bvij[13][18] = 0.975187766
gvij[13][18] = 1.171714677
btij[13][18] = 0.973091413
gtij[13][18] = 1.103693489

# C10H22-He
bvij[13][19] = 1
gvij[13][19] = 1
btij[13][19] = 1
gtij[13][19] = 1

# C10H22-Ar
bvij[13][20] = 1
gvij[13][20] = 1
btij[13][20] = 1
gtij[13][20] = 1

# H2-O2
bvij[14][15] = 1
gvij[14][15] = 1
btij[14][15] = 1
gtij[14][15] = 1

# H2-CO
bvij[14][16] = 1
gvij[14][16] = 1.121416201
btij[14][16] = 1
gtij[14][16] = 1.377504607

# H2-H2O
bvij[14][17] = 1
gvij[14][17] = 1
btij[14][17] = 1
gtij[14][17] = 1

# H2-H2S
bvij[14][18] = 1
gvij[14][18] = 1
btij[14][18] = 1
gtij[14][18] = 1

# H2-He
bvij[14][19] = 1
gvij[14][19] = 1
btij[14][19] = 1
gtij[14][19] = 1

# H2-Ar
bvij[14][20] = 1
gvij[14][20] = 1
btij[14][20] = 1
gtij[14][20] = 1

# O2-CO
bvij[15][16] = 1
gvij[15][16] = 1
btij[15][16] = 1
gtij[15][16] = 1

# O2-H2O
bvij[15][17] = 1
gvij[15][17] = 1.143174289
btij[15][17] = 1
gtij[15][17] = 0.964767932

# O2-H2S
bvij[15][18] = 1
gvij[15][18] = 1
btij[15][18] = 1
gtij[15][18] = 1

# O2-He
bvij[15][19] = 1
gvij[15][19] = 1
btij[15][19] = 1
gtij[15][19] = 1

# O2-Ar
bvij[15][20] = 0.999746847
gvij[15][20] = 0.993907223
btij[15][20] = 1.000023103
gtij[15][20] = 0.990430423

# CO-H2O
bvij[16][17] = 1
gvij[16][17] = 1
btij[16][17] = 1
gtij[16][17] = 1

# CO-H2S
bvij[16][18] = 0.795660392
gvij[16][18] = 1.101731308
btij[16][18] = 1.025536736
gtij[16][18] = 1.022749748

# CO-He
bvij[16][19] = 1
gvij[16][19] = 1
btij[16][19] = 1
gtij[16][19] = 1

# CO-Ar
bvij[16][20] = 1
gvij[16][20] = 1.159720623
btij[16][20] = 1
gtij[16][20] = 0.954215746

# H2O-H2S
bvij[17][18] = 1
gvij[17][18] = 1.014832832
btij[17][18] = 1
gtij[17][18] = 0.940587083

# H2O-He
bvij[17][19] = 1
gvij[17][19] = 1
btij[17][19] = 1
gtij[17][19] = 1

# H2O-Ar
bvij[17][20] = 1
gvij[17][20] = 1.038993495
btij[17][20] = 1
gtij[17][20] = 1.070941866

# H2S-He
bvij[18][19] = 1
gvij[18][19] = 1
btij[18][19] = 1
gtij[18][19] = 1

# H2S-Ar
bvij[18][20] = 1
gvij[18][20] = 1
btij[18][20] = 1
gtij[18][20] = 1

# He-Ar
bvij[19][20] = 1
gvij[19][20] = 1
btij[19][20] = 1
gtij[19][20] = 1

for i in range(MaxFlds):
    bvij[i][i] = 1
    btij[i][i] = 1
    gvij[i][i] = 1 / Dc[i]
    gtij[i][i] = Tc[i]
    for j in range(i + 1, MaxFlds):
        gvij[i][j] = gvij[i][j] * bvij[i][j] * (Vc3[i] + Vc3[j]) ** 3
        gtij[i][j] = gtij[i][j] * btij[i][j] * Tc2[i] * Tc2[j]
        bvij[i][j] = bvij[i][j] ** 2
        btij[i][j] = btij[i][j] ** 2

for i in range(MaxMdl):
    for j in range(MaxTrmM):
        gijk[i][j] = -cijk[i][j] * (eijk[i][j]**2) + bijk[i][j] * gijk[i][j]
        eijk[i][j] = 2 * cijk[i][j] * eijk[i][j] - bijk[i][j]
        cijk[i][j] = -cijk[i][j]

# Ideal gas terms
T0 = 298.15
d0 = 101.325 / RGERG / T0
for i in range(MaxFlds):
    n0i[i][2] = n0i[i][2] - 1
    n0i[i][1] = n0i[i][1] + T0
    for j in range(7):
        n0i[i][j] = Rsr * n0i[i][j]
    n0i[i][1] = n0i[i][1] - T0
    n0i[i][0] = n0i[i][0] - math.log(d0)
